// Copyright (c) 1999-2000 David Muse
// See the COPYING file for more information

#ifndef __CYGWIN__

#include <config.h>

#include <iostream.h>

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif
#include <stdlib.h>
#include <sys/stat.h>

#include <rudiments/semaphoreset.h>

#ifndef HAVE_SEMUN
union semun {
	int	val;
	struct	semid_ds	*buf;
	ushort	*array;
};
#endif

semaphoreset::semaphoreset() {

	showerrors=0;

	waitop=NULL;

	created=0;

	semid=-1;

	passwdent=NULL;
	groupent=NULL;
}

semaphoreset::~semaphoreset() {

	if (waitop) {
		for (int i=0; i<semcount; i++) {
			delete[] waitop[i];
			delete[] waitwithundoop[i];
			delete[] signalop[i];
			delete[] signalwithundoop[i];
		}
		delete[] waitop;
		delete[] waitwithundoop;
		delete[] signalop;
		delete[] signalwithundoop;
	}

	if (created) {
		forceRemove();
	}

	if (passwdent) {
		delete passwdent;
	}

	if (groupent) {
		delete groupent;
	}
}

int	semaphoreset::create(key_t key, mode_t permissions, 
					int semcount, int *values) {

	this->semcount=semcount;

	// create the semaphore
	if ((semid=semget(key,semcount,IPC_CREAT|IPC_EXCL|permissions))>=0) {

		// if creation succeeded, initialize the semaphore
		if (values) {
			for (int i=0; i<semcount; i++) {
				setValue(i,values[i]);
			}
		}

		// mark for removal
		created=1;

		createOperations();
		return 1;
	}

	// failure
	if (showerrors) {
		cerr << "error: couldn't create semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::attach(key_t key, int semcount) {

	this->semcount=semcount;

	if ((semid=semget(key,semcount,0))>=0) {
		createOperations();
		return 1;
	}

	// failure
	if (showerrors) {
		cerr << "error: couldn't attach to semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::createOrAttach(key_t key, mode_t permissions, 
						int semcount, int *values) {

	this->semcount=semcount;

	// create the semaphore
	if ((semid=semget(key,semcount,IPC_CREAT|IPC_EXCL|permissions))>=0) {

		// if creation succeeded, initialize the semaphore
		if (values) {
			for (int i=0; i<semcount; i++) {
				setValue(i,values[i]);
			}
		}

		// mark for removal
		created=1;
		
	} else if (!(errno==EEXIST && 
				(semid=semget(key,semcount,permissions))>=0)) {

		// failure
		if (showerrors) {
			cerr << "error: couldn't create or access semaphore: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}
	createOperations();
	return 1;
}

void	semaphoreset::dontRemove() {
	created=0;
}

int	semaphoreset::forceRemove() {
	if (semid==-1) {
		return 0;
	}
	semun	semctlun;
	if (!semctl(semid,0,IPC_RMID,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't remove semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

void	semaphoreset::showErrors() {
	showerrors=1;
}

void	semaphoreset::dontShowErrors() {
	showerrors=0;
}

void	semaphoreset::createOperations() {

	waitop=new sembuf *[semcount];
	waitwithundoop=new sembuf *[semcount];
	signalop=new sembuf *[semcount];
	signalwithundoop=new sembuf *[semcount];

	for (int i=0; i<semcount; i++) {

		// wait without undo
		waitop[i]=new sembuf[1];
		waitop[i][0].sem_num=(short)i;
		waitop[i][0].sem_op=-1;
		waitop[i][0].sem_flg=0;

		// wait with undo
		waitwithundoop[i]=new sembuf[1];
		waitwithundoop[i][0].sem_num=(short)i;
		waitwithundoop[i][0].sem_op=-1;
		waitwithundoop[i][0].sem_flg=SEM_UNDO;

		// signal without undo
		signalop[i]=new sembuf[1];
		signalop[i][0].sem_num=(short)i;
		signalop[i][0].sem_op=1;
		signalop[i][0].sem_flg=0;

		// signal with undo
		signalwithundoop[i]=new sembuf[1];
		signalwithundoop[i][0].sem_num=(short)i;
		signalwithundoop[i][0].sem_op=1;
		signalwithundoop[i][0].sem_flg=SEM_UNDO;
	}
}

int	semaphoreset::getId() {
	return semid;
}

int	semaphoreset::wait(int index) {
	if (semid==-1) {
		return 0;
	}
	if (!semop(semid,waitop[index],1)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't wait on semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::waitWithUndo(int index) {
	if (semid==-1) {
		return 0;
	}
	if (!semop(semid,waitwithundoop[index],1)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't wait with undo on semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::signal(int index) {
	if (semid==-1) {
		return 0;
	}
	if (!semop(semid,signalop[index],1)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't signal on semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::signalWithUndo(int index) {
	if (semid==-1) {
		return 0;
	}
	if (!semop(semid,signalwithundoop[index],1)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't signal with undo on semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::getValue(int index) {
	if (semid==-1) {
		return 0;
	}
	semun	semctlun;
	if (!semctl(semid,index,GETVAL,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't get value of semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::setValue(int index, int value) {
	if (semid==-1) {
		return 0;
	}
	semun	semctlun;
	semctlun.val=value;
	if (!semctl(semid,index,SETVAL,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set value of semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::getWaitingForZero(int index) {
	if (semid==-1) {
		return 0;
	}
	semun	semctlun;
	if (!semctl(semid,index,GETZCNT,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't get the number of processes " << endl;
		cerr << "       waiting for zero on semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::getWaitingForIncrement(int index) {
	if (semid==-1) {
		return 0;
	}
	semun	semctlun;
	if (!semctl(semid,index,GETNCNT,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't get the number of processes " << endl;
		cerr << "       waiting for increment on semaphore ";
		cerr << index << " : " << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::setUserName(char *username) {
	if (semid==-1) {
		return 0;
	}
	if (passwdent) {
		delete passwdent;
	}
	passwdent=getpwnam(username);
	if (!passwdent) {
		return 0;
	}
	return setUserId(passwdent->pw_uid);
}

int	semaphoreset::setUserId(ushort uid) {
	if (semid==-1) {
		return 0;
	}
	semun		semctlun;
	semid_ds	setds;
	setds.sem_perm.uid=uid;
	semctlun.buf=&setds;
	if (!semctl(semid,0,IPC_SET,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set user id for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::setGroupName(char *groupname) {
	if (semid==-1) {
		return -1;
	}
	if (groupent) {
		delete groupent;
	}
	groupent=getgrnam(groupname);
	if (!groupent) {
		return -1;
	}
	return setGroupId(groupent->gr_gid);
}

int	semaphoreset::setGroupId(ushort gid) {
	if (semid==-1) {
		return 0;
	}
	semun		semctlun;
	semid_ds	setds;
	setds.sem_perm.gid=gid;
	semctlun.buf=&setds;
	if (!semctl(semid,0,IPC_SET,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set group id for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	semaphoreset::setPermissions(mode_t permissions) {
	if (semid==-1) {
		return 0;
	}
	semun		semctlun;
	semid_ds	setds;
	setds.sem_perm.mode=permissions;
	semctlun.buf=&setds;
	if (!semctl(semid,0,IPC_SET,semctlun)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set permissions for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

char	*semaphoreset::getUserName() {
	if (semid==-1) {
		return NULL;
	}
	semun		semctlun;
	semid_ds	getds;
	semctlun.buf=&getds;
	if (!semctl(semid,0,IPC_STAT,semctlun)) {
		if (passwdent) {
			delete passwdent;
		}
		passwdent=getpwuid(getds.sem_perm.uid);
		return passwdent->pw_name;
	}
	if (showerrors) {
		cerr << "error: couldn't get user name for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return NULL;
}

unsigned short	semaphoreset::getUserId() {
	if (semid==-1) {
		return 0;
	}
	semun		semctlun;
	semid_ds	getds;
	semctlun.buf=&getds;
	if (!semctl(semid,0,IPC_STAT,semctlun)) {
		return (short)getds.sem_perm.uid;
	}
	if (showerrors) {
		cerr << "error: couldn't get user id for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

char	*semaphoreset::getGroupName() {
	if (semid==-1) {
		return NULL;
	}
	semun		semctlun;
	semid_ds	getds;
	semctlun.buf=&getds;
	if (!semctl(semid,0,IPC_STAT,semctlun)) {
		if (groupent) {
			delete groupent;
		}
		groupent=getgrgid(getds.sem_perm.gid);
		return groupent->gr_name;
	}
	if (showerrors) {
		cerr << "error: couldn't get group name for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return NULL;
}

unsigned short	semaphoreset::getGroupId() {
	if (semid==-1) {
		return 0;
	}
	semun		semctlun;
	semid_ds	getds;
	semctlun.buf=&getds;
	if (!semctl(semid,0,IPC_STAT,semctlun)) {
		return (short)getds.sem_perm.gid;
	}
	if (showerrors) {
		cerr << "error: couldn't get group id for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

mode_t	semaphoreset::getPermissions() {
	if (semid==-1) {
		return 0;
	}
	semun		semctlun;
	semid_ds	getds;
	semctlun.buf=&getds;
	if (!semctl(semid,0,IPC_STAT,semctlun)) {
		return getds.sem_perm.mode;
	}
	if (showerrors) {
		cerr << "error: couldn't get permissions for semaphore: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

#endif
