// Copyright (c) 1999-2000 David Muse
// See the COPYING file for more information

#include <config.h>

#include <iostream.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <rudiments/serversocket.h>

serversocket::serversocket() : genericsocket() {
	serverport=0;
}

int	serversocket::initInetPort(unsigned short int port) {

	// check for an initialized socket
	if (sock>-1) {
		if (showerrors) {
			cerr << "error: socket is already initialized" << endl;
		}
		return 0;
	}

	// this is an inet socket
	inetsocket=1;

	// init the socket structure
	bzero((void *)&serversin,sizeof(serversin));
	serversin.sin_family=AF_INET;
	serversin.sin_port=htons(port);
	serversin.sin_addr.s_addr=htonl(INADDR_ANY);

	// init the port
	serverport=port;

	// create the socket
	if ((sock=socket(AF_INET,SOCK_STREAM,0))==-1) {
		if (showerrors) {
			cerr << "error: inet socket creation failed: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	return 1;
}

int	serversocket::initUnixPort(char *port, mode_t mode) {

	// check for an initialized socket
	if (sock>-1) {
		if (showerrors) {
			cerr << "error: socket is already initialized" << endl;
		}
		return 0;
	}

	// this is a unix socket
	unixsocket=1;

	// if a null or blank port was specified, return an error
	if (!port || (port && !port[0])) {
		return 0;
	}

	// init the socket structure
	unlink(port);
	bzero((void *)&serversun,sizeof(serversun));
	serversun.sun_family=AF_UNIX;
	strcpy(serversun.sun_path,port);

	// create the socket
	if ((sock=socket(AF_UNIX,SOCK_STREAM,0))==-1) {
		if (showerrors) {
			cerr << "error: unix socket creation failed: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	// set the mode of the file
	chmod(serversun.sun_path,mode);

	return 1;
}

int	serversocket::lingerOnClose(int timeout) {

	// check for an uninitialized socket
	if (sock==-1) {
		if (showerrors) {
			cerr << "error: socket is uninitialized";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	// set the linger socket option
	struct	linger	ling;
	ling.l_onoff=1;
	ling.l_linger=timeout;
	if (setsockopt(sock,SOL_SOCKET,SO_LINGER,
			(void *)&ling,sizeof(struct linger))==-1) {
		if (showerrors) {
			cerr << "error: linger failed: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}
	return 1;
}

int	serversocket::reuseAddresses() {

	// check for an uninitialized socket
	if (sock==-1) {
		if (showerrors) {
			cerr << "error: socket is uninitialized";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	// set the reuse addresses socket option
	int	one=1;
	if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,
			(void *)&one,(socklen_t)sizeof(int))==-1) {
		if (showerrors) {
			cerr << "error: reuse addresses failed: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}
	return 1;
}

int	serversocket::bindSocket() {

	// check for an uninitialized socket
	if (sock==-1) {
		if (showerrors) {
			cerr << "error: socket is uninitialized";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	// bind the socket
	if ((inetsocket && bind(sock,(struct sockaddr *)&serversin,
					sizeof(serversin))==-1) ||
		(unixsocket && bind(sock,(struct sockaddr *)&serversun,
					sizeof(serversun))==-1)) {
		if (showerrors) {
			cerr << "error: bind failed: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	// get the actual port number if an arbitrary port was requested
	if (inetsocket && !serverport) {
		struct	sockaddr_in	testsin;
		socklen_t	len=sizeof(testsin);
		bzero((void *)&testsin,len);
		getsockname(sock,(struct sockaddr *)&testsin,&len);
		serverport=(unsigned short int)ntohs(testsin.sin_port);
	}
	return 1;
}

int	serversocket::listenForClientConnection(int backlog) {

	// check for an uninitialized socket
	if (sock==-1) {
		if (showerrors) {
			cerr << "error: socket is uninitialized";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}

	// listen on the socket
	if (listen(sock,backlog)==-1) {
		if (showerrors) {
			cerr << "error: inet listen failed: ";
			cerr << strerror(errno) << endl;
		}
		return 0;
	}
	return 1;
}

int	serversocket::listenOnInetPort(unsigned short int port, int backlog) {
	return initInetPort(port) && bindSocket() && 
			listenForClientConnection(backlog);
}

int	serversocket::listenOnUnixPort(char *port, mode_t mode, int backlog) {
	return initUnixPort(port,mode) && bindSocket() && 
			listenForClientConnection(backlog);
}

genericsocket	*serversocket::acceptClientConnection() {

	// check for an uninitialized socket
	if (sock==-1) {
		if (showerrors) {
			cerr << "error: socket is uninitialized";
			cerr << strerror(errno) << endl;
		}
		return NULL;
	}

	// accept on the socket
	socklen_t	clientsinlen=sizeof(clientsin);
	socklen_t	clientsunlen=sizeof(clientsun);
	int		clientsock;
	if ((inetsocket && 
		(clientsock=accept(sock,(struct sockaddr *)&clientsin,
						&clientsinlen))==-1) ||
		(unixsocket && 
		(clientsock=accept(sock,(struct sockaddr *)&clientsun,
						&clientsunlen))==-1)) {
		if (showerrors) {
			cerr << "error: accept failed: ";
			cerr << strerror(errno) << endl;
		}
		return NULL;
	}

	// create a server to return
	return new genericsocket(clientsock);
}

unsigned short int	serversocket::getPort() {
	return serverport;
}

int	serversocket::getClientAddress(genericsocket *sock, 
							char *buffer) {

	// if the socket is not an inet socket, set the ip to localhost
	if (sock->isUnixSocket()) {
		strcpy(buffer,"127.0.0.1");
		return 1;
	}

	// get the peer name then convert the address to a string
	struct sockaddr_in	client;
	int			len=sizeof(client);
	if (getpeername(sock->getDescriptor(),
			(struct sockaddr *)&client,(socklen_t *)&len)) {
		return 0;
	}
	strcpy(buffer,inet_ntoa(client.sin_addr));
	return 1;
}
