// Copyright (c) 1999-2000 David Muse
// See the COPYING file for more information

#ifndef __CYGWIN__

#include <config.h>

#include <iostream.h>

#include <sys/types.h>
#include <errno.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif
#include <stdlib.h>
#include <sys/stat.h>

#include <rudiments/sharedmemory.h>

sharedmemory::sharedmemory() {
	showerrors=0;
	created=0;
	shmptr=NULL;
	passwdent=NULL;
	groupent=NULL;
	shmid=-1;
}

sharedmemory::~sharedmemory() {
	if (created) {
		forceRemove();
	}
	if (passwdent) {
		delete passwdent;
	}
	if (groupent) {
		delete groupent;
	}
}

int	sharedmemory::create(key_t key, int size, mode_t permissions) {

	// create the shared memory segment
	if ((shmid=shmget(key,size,IPC_CREAT|IPC_EXCL|permissions))>=0) {

		// mark for removal
		created=1;

		// attach to the segment
		shmptr=shmat(shmid,0,0);

		// init the segment to zero's
		memset((void *)shmptr,0,size);
		return 1;
	}

	// failure
	if (showerrors) {
		cerr << "error: couldn't create shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	sharedmemory::attach(key_t key) {

	if ((shmid=shmget(key,0,0))>=0) {

		// attach to the segment
		shmptr=shmat(shmid,0,0);
		return 1;
	}

	// failure
	if (showerrors) {
		cerr << "error: couldn't attach to shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	sharedmemory::createOrAttach(key_t key, int size, mode_t permissions) {

	// create the shared memory segment
	if ((shmid=shmget(key,size,IPC_CREAT|IPC_EXCL|permissions))>=0) {

		// mark for removal
		created=1;

		// attach to the segment
		shmptr=shmat(shmid,0,0);

		// init the segment to zero's
		memset((void *)shmptr,0,size);
		return 1;
		
	} else if (errno==EEXIST && (shmid=shmget(key,0,permissions))>=0) {

		// attach to the segment
		shmptr=shmat(shmid,0,0);
		return 1;

	}

	// failure
	if (showerrors) {
		cerr << "error: couldn't create or access shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

void	sharedmemory::dontRemove() {
	created=0;
}

int	sharedmemory::forceRemove() {
	if (shmid==-1) {
		return 0;
	}
	if (!shmctl(shmid,IPC_RMID,NULL)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't remove shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

void	sharedmemory::showErrors() {
	showerrors=1;
}

void	sharedmemory::dontShowErrors() {
	showerrors=0;
}

int	sharedmemory::getId() {
	return shmid;
}

void	*sharedmemory::getPointer() {
	return shmptr;
}

int	sharedmemory::setUserName(char *username) {
	if (shmid==-1) {
		return 0;
	}
	if (passwdent) {
		delete passwdent;
	}
	passwdent=getpwnam(username);
	if (!passwdent) {
		return 0;
	}
	return setUserId(passwdent->pw_uid);
}

int	sharedmemory::setGroupName(char *groupname) {
	if (shmid==-1) {
		return 0;
	}
	if (groupent) {
		delete groupent;
	}
	groupent=getgrnam(groupname);
	if (!groupent) {
		return 0;
	}
	return setGroupId(groupent->gr_gid);
}

int	sharedmemory::setUserId(ushort uid) {
	if (shmid==-1) {
		return 0;
	}
	shmid_ds	setds;
	setds.shm_perm.uid=uid;
	if (!shmctl(shmid,IPC_SET,&setds)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set user id for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	sharedmemory::setGroupId(ushort gid) {
	if (shmid==-1) {
		return 0;
	}
	shmid_ds	setds;
	setds.shm_perm.gid=gid;
	if (!shmctl(shmid,IPC_SET,&setds)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set group id for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

int	sharedmemory::setPermissions(mode_t permissions) {
	if (shmid==-1) {
		return 0;
	}
	shmid_ds	setds;
	setds.shm_perm.mode=permissions;
	if (!shmctl(shmid,IPC_SET,&setds)) {
		return 1;
	}
	if (showerrors) {
		cerr << "error: couldn't set permissions for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

char	*sharedmemory::getUserName() {
	if (shmid==-1) {
		return NULL;
	}
	shmid_ds	getds;
	if (!shmctl(shmid,IPC_STAT,&getds)) {
		if (passwdent) {
			delete passwdent;
		}
		passwdent=getpwuid(getds.shm_perm.uid);
		return passwdent->pw_name;
	}
	if (showerrors) {
		cerr << "error: couldn't get user name for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return NULL;
}

char	*sharedmemory::getGroupName() {
	if (shmid==-1) {
		return NULL;
	}
	shmid_ds	getds;
	if (!shmctl(shmid,IPC_STAT,&getds)) {
		if (groupent) {
			delete groupent;
		}
		groupent=getgrgid(getds.shm_perm.gid);
		return groupent->gr_name;
	}
	if (showerrors) {
		cerr << "error: couldn't get user name for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return NULL;
}

unsigned short	sharedmemory::getUserId() {
	if (shmid==-1) {
		return 0;
	}
	shmid_ds	getds;
	if (!shmctl(shmid,IPC_STAT,&getds)) {
		return getds.shm_perm.uid;
	}
	if (showerrors) {
		cerr << "error: couldn't get user id for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

unsigned short	sharedmemory::getGroupId() {
	if (shmid==-1) {
		return 0;
	}
	shmid_ds	getds;
	if (!shmctl(shmid,IPC_STAT,&getds)) {
		return getds.shm_perm.gid;
	}
	if (showerrors) {
		cerr << "error: couldn't get group id for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

mode_t	sharedmemory::getPermissions() {
	if (shmid==-1) {
		return 0;
	}
	shmid_ds	getds;
	if (!shmctl(shmid,IPC_STAT,&getds)) {
		return getds.shm_perm.mode;
	}
	if (showerrors) {
		cerr << "error: couldn't get permissions for shared memory: ";
		cerr << strerror(errno) << endl;
	}
	return 0;
}

#endif
